/*###################################################################*/
/*##                         GQmpeg 0.4.5                          ##*/
/*##    This software is Copyright (C) 1998, 1999 by John Ellis.   ##*/
/*## This software falls under the GNU Public License. Please read ##*/
/*##              the COPYING file for more information            ##*/
/*###################################################################*/

#include "gqmpeg.h"

#define DIALOG_SAVE 1
#define DIALOG_LOAD 2
#define DIALOG_APPEND 3

typedef struct _FileDialog FileDialog;
struct _FileDialog
{
	GtkWidget *dialog;
	GtkWidget *dir_entry;
	GtkWidget *file_entry;
	gchar *src_path;
	gchar *src_dir;
	gchar *src_file;
	gchar *dest_path;
};

static void warning_dialog_close(GtkWidget *w, gpointer data);
static void warning_dialog_destroy(GtkWidget *w, GdkEvent *event, gpointer data);
static FileDialog *new_dialog_data(gchar *fn);
static void file_manage_close(GtkWidget *widget, gpointer data);
static void file_manage_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static void file_rename_confirmed(GtkWidget *widget, gpointer data);
static void file_delete_confirmed(GtkWidget *widget, gpointer data);
static void file_move_or_copy_confirmed(FileDialog *fd, gint move, gint overwrite);
static void file_move_overwrite_confirmed(GtkWidget *widget, gpointer data);
static void file_copy_overwrite_confirmed(GtkWidget *widget, gpointer data);
static void file_move_confirmed(GtkWidget *widget, gpointer data);
static void file_copy_confirmed(GtkWidget *widget, gpointer data);
static void file_move_confirmed_by_enter(gchar *fn, gpointer data);
static void file_copy_confirmed_by_enter(gchar *fn, gpointer data);
static void file_move_or_copy_dialog(gchar *fn, gint move);
static void playlist_save_confirmed(FileDialog *fd, gint overwrite);
static void playlist_save_pressed(GtkWidget *widget, gpointer data);
static void playlist_save_overwrite_pressed(GtkWidget *widget, gpointer data);
static void playlist_load_confirmed(GtkWidget *widget, gpointer data);
static void playlist_append_confirmed(GtkWidget *widget, gpointer data);
static void playlist_save_confirmed_by_enter(gchar *fn, gpointer data);
static void playlist_load_confirmed_by_enter(gchar *fn, gpointer data);
static void playlist_append_confirmed_by_enter(gchar *fn, gpointer data);
static void playlist_save_or_load_dialog(gchar *fn, gint type);

static void warning_dialog_close(GtkWidget *w, gpointer data)
{
	GtkWidget *warning_window = data;
	gtk_widget_destroy(warning_window);
}

static void warning_dialog_destroy(GtkWidget *w, GdkEvent *event, gpointer data)
{
	warning_dialog_close(NULL, data);
}

void warning_dialog(gchar *title, gchar *message)
{
	GtkWidget *warning_window;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *button;

	warning_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT (warning_window), "delete_event",(GtkSignalFunc) warning_dialog_destroy, warning_window);
	gtk_window_set_policy (GTK_WINDOW(warning_window), FALSE, FALSE, TRUE);
	gtk_window_set_title (GTK_WINDOW (warning_window), title);
	gtk_container_border_width (GTK_CONTAINER (warning_window), 10);

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(warning_window),vbox);
	gtk_widget_show(vbox);

	label = gtk_label_new(message);
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	hbox = gtk_hbox_new(TRUE, 5);
	gtk_container_add(GTK_CONTAINER(vbox),hbox);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("     Ok     ");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) warning_dialog_close, warning_window);
	gtk_box_pack_start(GTK_BOX(hbox),button,FALSE,FALSE,0);
	gtk_widget_show(button);
	
	gtk_widget_show(warning_window);
}

static FileDialog *new_dialog_data(gchar *fn)
{
	FileDialog *fd;
	gchar *buf;
	gchar *ptr;

	fd = g_new0(FileDialog, 1);
	fd->src_path = g_strdup(fn);
	fd->src_file = g_strdup(filename_from_path(fn));
	buf = g_strdup(fn);
	ptr = filename_from_path(buf);
	if (ptr > buf) ptr--;
	ptr[0] = '\0';
	fd->src_dir = g_strdup(buf);
	g_free(buf);

	return fd;
}

static void file_manage_close(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	if(fd->dialog) gtk_widget_destroy(fd->dialog);
	g_free(fd->src_path);
	g_free(fd->src_dir);
	g_free(fd->src_file);
	g_free(fd->dest_path);
	g_free(fd);
}

static void file_manage_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	file_manage_close(NULL, data);
}

static void file_rename_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	gint old_status = status;
	gint song_num;

	fd->dest_path = g_strconcat(fd->src_dir, "/", gtk_entry_get_text(GTK_ENTRY(fd->file_entry)), NULL);

	song_num = playlist_get_index(fd->src_path);
	if (song_num == playlist_song && status == STATUS_PLAY)
		{
		stop_playing_file();
		status = STATUS_PAUSE;
		}

	if ( (rename (fd->src_path, fd->dest_path) < 0) )
		{
		/* rename failed */
		gchar *buf;
		buf = g_strconcat("Unable to rename the file:\n", fd->src_file, NULL);
		warning_dialog("Rename failed", buf);
		g_free(buf);

		if (old_status != status && old_status == STATUS_PLAY) start_playing_file();
		}
	else
		{
		/* update_lists */
		playlist_replace_song(fd->src_path, fd->dest_path);
		refresh_dir();
		}
	if (old_status == STATUS_PLAY) start_playing_file();

	file_manage_close(NULL, fd);
}

void file_rename_dialog(gchar *fn)
{
	FileDialog *fd;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *button;

	if (!fn) return;

	fd = new_dialog_data(fn);

	fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event", (GtkSignalFunc)file_manage_delete_cb, fd);
	gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
	gtk_window_set_title (GTK_WINDOW (fd->dialog), "Rename File");
	gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
	gtk_widget_show(vbox);

	label = gtk_label_new("Rename the file:");
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_label_new(fd->src_file);
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_label_new("to:");
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	fd->file_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(fd->file_entry),fd->src_file);
	gtk_box_pack_start(GTK_BOX(vbox), fd->file_entry, FALSE,FALSE,0);
	gtk_widget_grab_focus(fd->file_entry);
	gtk_widget_show(fd->file_entry);
	gtk_entry_select_region(GTK_ENTRY(fd->file_entry), 0, strlen(fd->src_file));

	label = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	hbox = gtk_hbox_new(TRUE,0);
	gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("Rename");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_rename_confirmed, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	gtk_widget_show(fd->dialog);
}

static void file_delete_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;

	if ( (unlink (fd->src_path) < 0) )
		{
		/* delete failed */
		gchar *buf;
		buf = g_strconcat("Unable to delete the file:\n", fd->src_file, NULL);
		warning_dialog("Delete failed",buf);
		g_free(buf);
		}
	else
		{
		/* update lists */
		playlist_remove_song(fd->src_path, 0);
		refresh_dir();
		}
	file_manage_close(NULL, fd);
}

void file_delete_dialog(gchar *fn)
{
	FileDialog *fd;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *button;

	if (!fn) return;

	fd = new_dialog_data(fn);

	fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event",(GtkSignalFunc) file_manage_delete_cb, fd);
	gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
	gtk_window_set_title (GTK_WINDOW (fd->dialog), "Delete File");
	gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
	gtk_widget_show(vbox);

	label = gtk_label_new("About to delete the file named:");
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_label_new(fd->src_file);
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	hbox = gtk_hbox_new(TRUE,0);
	gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("Delete");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_delete_confirmed, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_grab_focus(button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	gtk_widget_show(fd->dialog);
}

static void file_move_or_copy_confirmed(FileDialog *fd, gint move, gint overwrite)
{
	if (!overwrite)
		fd->dest_path = remove_trailing_slash(gtk_entry_get_text(GTK_ENTRY(fd->dir_entry)));

	if (isdir(fd->dest_path))
		{
		gchar *buf;
		buf = g_strconcat(fd->dest_path, "/", fd->src_file, NULL);
		g_free(fd->dest_path);
		fd->dest_path = buf;
		}

	if (strcmp(fd->src_path, fd->dest_path) == 0)
		{
		gchar *buf;
		buf = g_strconcat("Source and destination are the same for:\n" ,fd->src_path, NULL);
		if (move)
			warning_dialog("Cannot move a file to itself",buf);
		else
			warning_dialog("Cannot copy a file to itself",buf);
		file_manage_close(NULL, fd);
		g_free(buf);
		return;
		}

	if (!overwrite && isfile(fd->dest_path))
		{
		GtkWidget *vbox;
		GtkWidget *hbox;
		GtkWidget *label;
		GtkWidget *button;

		gtk_widget_destroy(fd->dialog);
		fd->dialog = NULL;

		fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event",(GtkSignalFunc) file_manage_delete_cb, fd);
		gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Overwrite File");
		gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

		vbox = gtk_vbox_new(FALSE,5);
		gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
		gtk_widget_show(vbox);

		label = gtk_label_new("Overwrite file:");
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		label = gtk_label_new(fd->dest_path);
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		label = gtk_label_new("with:");
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		label = gtk_label_new(fd->src_path);
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		hbox = gtk_hbox_new(TRUE,0);
		gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
		gtk_widget_show(hbox);

		button = gtk_button_new_with_label("Overwrite");
		if (move)
			gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_move_overwrite_confirmed, fd);
		else
			gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_copy_overwrite_confirmed, fd);
		gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
		gtk_widget_show(button);

		button = gtk_button_new_with_label("Cancel");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
		gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
		gtk_widget_show(button);

		gtk_widget_show(fd->dialog);

		return;
		}

	if (move)
		{
		if (!move_file(fd->src_path, fd->dest_path))
			{
			gchar *buf;
			buf = g_strconcat("Unable to move the file:\n", fd->src_file, NULL);
			warning_dialog("Move failed",buf);
			g_free(buf);
			}
		else
			{
			/* update lists */
			playlist_replace_song(fd->src_path, fd->dest_path);
			refresh_dir();
			}
		}
	else
		{
		if (!copy_file(fd->src_path, fd->dest_path))
			{
			gchar *buf;
			buf = g_strconcat("Unable to copy the file:\n", fd->src_file, NULL);
			warning_dialog("Copy failed",buf);
			g_free(buf);
			}
		}

	file_manage_close(NULL, fd);
}

static void file_move_overwrite_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, TRUE, TRUE);
}

static void file_copy_overwrite_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, FALSE, TRUE);
}

static void file_move_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, TRUE, FALSE);
}

static void file_copy_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, FALSE, FALSE);
}

static void file_move_confirmed_by_enter(gchar *fn, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, TRUE, FALSE);
}

static void file_copy_confirmed_by_enter(gchar *fn, gpointer data)
{
	FileDialog *fd = data;
	file_move_or_copy_confirmed(fd, FALSE, FALSE);
}

static void file_move_or_copy_dialog(gchar *fn, gint move)
{
	FileDialog *fd;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *button;
	GtkWidget *list;

	if (!fn) return;

	fd = new_dialog_data(fn);

	fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(fd->dialog, 300, 250);
	gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event",(GtkSignalFunc) file_manage_delete_cb, fd);
	gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
	if (move)
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Move File");
	else
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Copy File");
	gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
	gtk_widget_show(vbox);

	if (move)
		label = gtk_label_new("Move file:");
	else
		label = gtk_label_new("Copy file:");
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_label_new(fd->src_file);
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	label = gtk_label_new("to:");
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	fd->dir_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(fd->dir_entry),fd->src_dir);
	gtk_box_pack_start(GTK_BOX(vbox), fd->dir_entry, FALSE,FALSE,0);
	gtk_widget_grab_focus(fd->dir_entry);
	gtk_widget_show(fd->dir_entry);
	gtk_entry_select_region(GTK_ENTRY(fd->dir_entry), 0, strlen(fd->src_dir));

	list = destination_widget_new(fd->src_dir, fd->dir_entry);
	gtk_box_pack_start(GTK_BOX(vbox),list,TRUE,TRUE,0);
	gtk_widget_show(list);

	label = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	hbox = gtk_hbox_new(TRUE,0);
	gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
	gtk_widget_show(hbox);

	if (move)
		{
		button = gtk_button_new_with_label("Move");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_move_confirmed, fd);
		tab_completion_add_to_entry(fd->dir_entry, file_move_confirmed_by_enter, fd);
		}
	else
		{
		button = gtk_button_new_with_label("Copy");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_copy_confirmed, fd);
		tab_completion_add_to_entry(fd->dir_entry, file_copy_confirmed_by_enter, fd);
		}
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	gtk_widget_show(fd->dialog);
}

void file_move_dialog(gchar *fn)
{
	file_move_or_copy_dialog(fn, TRUE);
}

void file_copy_dialog(gchar *fn)
{
	file_move_or_copy_dialog(fn, FALSE);
}

static void playlist_save_confirmed(FileDialog *fd, gint overwrite)
{
	if (isdir(fd->dest_path))
		{
		gchar *buf;
		buf = g_strconcat("specified path is a directory, playlists are files\n", fd->dest_path, NULL);
		warning_dialog("Invalid filename", buf);
		g_free(buf);
		g_free(fd->dest_path);
		fd->dest_path = buf;
		return;
		}

	if (!overwrite && isfile(fd->dest_path))
		{
		GtkWidget *vbox;
		GtkWidget *hbox;
		GtkWidget *label;
		GtkWidget *button;

		gtk_widget_destroy(fd->dialog);
		fd->dialog = NULL;

		fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event",(GtkSignalFunc) file_manage_delete_cb, fd);
		gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Overwrite File");
		gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

		vbox = gtk_vbox_new(FALSE,5);
		gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
		gtk_widget_show(vbox);

		label = gtk_label_new("Overwrite playlist file:");
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		label = gtk_label_new(fd->dest_path);
		gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
		gtk_widget_show(label);

		hbox = gtk_hbox_new(TRUE,0);
		gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
		gtk_widget_show(hbox);

		button = gtk_button_new_with_label("Overwrite");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) playlist_save_overwrite_pressed, fd);
		gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
		gtk_widget_show(button);

		button = gtk_button_new_with_label("Cancel");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
		gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
		gtk_widget_show(button);

		gtk_widget_show(fd->dialog);

		return;
		}

	if (playlist_save(fd->dest_path))
		{
		gchar *buf = g_strconcat("GQmpeg - ", filename_from_path(fd->dest_path), NULL);
		gtk_window_set_title (GTK_WINDOW (mainwindow), buf);
		g_free(buf);
		if (playlist_window)
			{
			buf = g_strconcat("GQmpeg Playlist - ", filename_from_path(fd->dest_path), NULL);
			gtk_window_set_title (GTK_WINDOW (playlist_window), buf);
			g_free(buf);
			refresh_dir();
			}
		if (playlist_pathname) g_free(playlist_pathname);
	        playlist_pathname = g_strdup(fd->dest_path);
        	if (playlist_filename) g_free(playlist_filename);
	        playlist_filename = g_strdup(filename_from_path(playlist_pathname));
		}
	else
		{
		gchar *buf = g_strconcat("Failed to save the playlist:\n", fd->dest_path, NULL);
		warning_dialog("Save Failed", buf);
		g_free(buf);
		}

	file_manage_close(NULL, fd);
}

static void playlist_save_pressed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	gchar *buf = gtk_entry_get_text(GTK_ENTRY(fd->dir_entry));

	if (strlen(buf) > 7 && strncasecmp(buf + (strlen(buf) - 7), ".gqmpeg", 7) == 0)
		fd->dest_path = g_strdup(buf);
	else
		fd->dest_path = g_strconcat(buf, ".gqmpeg", NULL);

	playlist_save_confirmed(fd, FALSE);

}

static void playlist_save_overwrite_pressed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	playlist_save_confirmed(fd, TRUE);
}

static void playlist_load_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	fd->dest_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(fd->dir_entry)));

	load_playlist_from_file(fd->dest_path);

	file_manage_close(NULL, fd);
}

static void playlist_append_confirmed(GtkWidget *widget, gpointer data)
{
	FileDialog *fd = data;
	fd->dest_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(fd->dir_entry)));

	append_playlist_from_file(fd->dest_path);

	file_manage_close(NULL, fd);
}

static void playlist_save_confirmed_by_enter(gchar *fn, gpointer data)
{
	FileDialog *fd = data;
	playlist_save_pressed(NULL, fd);
}

static void playlist_load_confirmed_by_enter(gchar *fn, gpointer data)
{
	FileDialog *fd = data;
	playlist_load_confirmed(NULL, fd);
}

static void playlist_append_confirmed_by_enter(gchar *fn, gpointer data)
{
	FileDialog *fd = data;
	playlist_append_confirmed(NULL, fd);
}

static void playlist_save_or_load_dialog(gchar *fn, gint type)
{
	FileDialog *fd;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *button;
	GtkWidget *list;

	if (!fn) return;

	fd = new_dialog_data(fn);

	fd->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(fd->dialog, 300, 350);
	gtk_signal_connect (GTK_OBJECT (fd->dialog), "delete_event",(GtkSignalFunc) file_manage_delete_cb, fd);
	gtk_window_set_policy (GTK_WINDOW(fd->dialog), FALSE, FALSE, TRUE);
	if (type == DIALOG_SAVE)
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Save Playlist");
	else if (type == DIALOG_LOAD)
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Load Playlist");
	else
		gtk_window_set_title (GTK_WINDOW (fd->dialog), "Append Playlist");

	gtk_container_border_width (GTK_CONTAINER (fd->dialog), 10);

	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_add (GTK_CONTAINER(fd->dialog),vbox);
	gtk_widget_show(vbox);

	if (type == DIALOG_SAVE)
		label = gtk_label_new("Save playlist as:");
	else if (type == DIALOG_LOAD)
		label = gtk_label_new("Load playlist from:");
	else
		label = gtk_label_new("Append playlist from:");

	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	fd->dir_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(fd->dir_entry),fd->src_path);
	gtk_box_pack_start(GTK_BOX(vbox), fd->dir_entry, FALSE,FALSE,0);
	gtk_widget_grab_focus(fd->dir_entry);
	gtk_widget_show(fd->dir_entry);
	gtk_entry_select_region(GTK_ENTRY(fd->dir_entry), 0, strlen(fd->src_path));

	list = destination_widget_new_with_files(fd->src_dir, ".gqmpeg", fd->dir_entry);
	gtk_box_pack_start(GTK_BOX(vbox),list,TRUE,TRUE,0);
	gtk_widget_show(list);

	label = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);
	gtk_widget_show(label);

	hbox = gtk_hbox_new(TRUE,0);
	gtk_box_pack_start(GTK_BOX(vbox),hbox,FALSE,FALSE,0);
	gtk_widget_show(hbox);

	if (type == DIALOG_SAVE)
		{
		button = gtk_button_new_with_label("Save");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) playlist_save_pressed, fd);
		tab_completion_add_to_entry(fd->dir_entry, playlist_save_confirmed_by_enter, fd);
		}
	else if (type == DIALOG_LOAD)
		{
		button = gtk_button_new_with_label("Load");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) playlist_load_confirmed, fd);
		tab_completion_add_to_entry(fd->dir_entry, playlist_load_confirmed_by_enter, fd);
		}
	else
		{
		button = gtk_button_new_with_label("Append");
		gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) playlist_append_confirmed, fd);
		tab_completion_add_to_entry(fd->dir_entry, playlist_append_confirmed_by_enter, fd);
		}

	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",(GtkSignalFunc) file_manage_close, fd);
	gtk_box_pack_start(GTK_BOX(hbox),button,TRUE,TRUE,10);
	gtk_widget_show(button);

	gtk_widget_show(fd->dialog);
}

void playlist_save_as_dialog(gchar *fn)
{
	if (playlist_count < 1) return;
	playlist_save_or_load_dialog(fn, DIALOG_SAVE);
}

void playlist_load_dialog(gchar *fn)
{
	playlist_save_or_load_dialog(fn, DIALOG_LOAD);
}

void playlist_append_dialog(gchar *fn)
{
	playlist_save_or_load_dialog(fn, DIALOG_APPEND);
}

